// -------------------------------------------
// pMini : a minimal OpenGL/GLUT application
// for point set rendering.
// Copyright (C) 2009 Tamy Boubekeur
// All rights reserved.
// -------------------------------------------

// -------------------------------------------
// Disclaimer: this code is dirty in the
// meaning that there is no attention paid to
// proper class attribute access, memory
// management or optimisation of any kind. It
// is designed for quick-and-dirty testing
// purpose.
// -------------------------------------------

#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <cstdio>
#include <cstdlib>

#include <algorithm>
#include <GL/glut.h>
#include <Vec3D.h>
#include "Camera.h"

using namespace std;

class Exception {
private:
    string msg;
public:
    inline Exception ( const string & msg) : msg (msg) {}
    virtual ~Exception () {}
    inline const string & getMessage () const { return msg; }
};

// -------------------------------------------
// Basic Mesh class
// -------------------------------------------

class Vertex {
public:
    inline Vertex () {}
    inline Vertex (const Vec3Df & p, const Vec3Df & n) : p (p), n (n) {}
    inline Vertex (const Vertex & v) : p (v.p), n (v.n) {}
    inline virtual ~Vertex () {}
    inline Vertex & operator= (const Vertex & v) {
        p = v.p;
        n = v.n;
        return (*this);
    }
    Vec3Df p;
    Vec3Df n;
};

class PointSet {
public:
    vector<Vertex> V;
    unsigned int glID;
    void loadPN (const string & filename);
    void centerAndScaleToUnit ();
    void buildGLList ();
};

static const unsigned int READ_BUFFER_SIZE = 1024;

void PointSet::loadPN (const string & filename) {
    unsigned int surfelSize = 6;
#ifdef _MSC_VER
    FILE * in;
    if (!fopen_s (&in, filename.c_str (), "rb"))
        throw Exception ("Bad PN file.");
#else
    FILE * in = fopen (filename.c_str (), "rb");
    if (in == NULL)
        throw Exception ("Bad PN file.");
#endif
    float * pn = new float[surfelSize*READ_BUFFER_SIZE];
    while (!feof (in)) {
        unsigned numOfVertex = fread (pn, 4, surfelSize*READ_BUFFER_SIZE, in);
        for (unsigned int i = 0; i < numOfVertex; i += surfelSize) 
            V.push_back (Vertex (Vec3Df (pn[i], pn[i+1], pn[i+2]), 
                                 Vec3Df (pn[i+3], pn[i+4], pn[i+5])));
        
        if (numOfVertex < surfelSize*READ_BUFFER_SIZE)
            break;
    }
    fclose (in);
    delete [] pn;
    centerAndScaleToUnit ();
    buildGLList ();
}

void PointSet::centerAndScaleToUnit () {
    Vec3Df c;
    for  (unsigned int i = 0; i < V.size (); i++)
        c += V[i].p;
    c /= V.size ();
    float maxD = Vec3Df::distance (V[0].p, c);
    for (unsigned int i = 0; i < V.size (); i++){
        float m = Vec3Df::distance (V[i].p, c);
        if (m > maxD)
            maxD = m;
    }
    for  (unsigned int i = 0; i < V.size (); i++)
        V[i].p = (V[i].p - c) / maxD;
}

void PointSet::buildGLList () {
    glID = glGenLists (1);
    glNewList (glID, GL_COMPILE);
    glBegin (GL_POINTS);
    for (unsigned int i = 0; i < V.size (); i++)  {
        const Vertex & v = V[i];
        glNormal3f (v.n[0], v.n[1], v.n[2]);
        glVertex3f (v.p[0], v.p[1], v.p[2]);
    }
    glEnd ();
    glEndList ();
}

// -------------------------------------------
// OpenGL/GLUT application code.
// -------------------------------------------

static GLint window;
static unsigned int SCREENWIDTH = 640;
static unsigned int SCREENHEIGHT = 480;
static Camera camera;
static bool mouseRotatePressed = false;
static bool mouseMovePressed = false;
static bool mouseZoomPressed = false;
static int lastX=0, lastY=0, lastZoom=0;
static unsigned int FPS = 0;
static bool fullScreen = false;

static PointSet ps;
static float pointSize = 2.0;

void printUsage () {
    cerr << endl 
         << "pMini: a minimal OpenGL/GLUT PointSet viewer" << endl
         << "Author : Tamy Boubekeur (http://www.telecom-paristech.fr/~boubek)" << endl << endl
         << "Usage : ./pmini [<file.pn>]" << endl
         << "PN format : binary file made of a list of 6 floats chunk:" << endl
         << "   x0,y0,z0,nx0,ny0,nz0, x1,y1,z1,nx1,ny1,nz1,..." << endl
         << "with {x,y,z} the position and {nx,ny,nz} the normal vector of each point sample" << endl
         << "Keyboard commands" << endl 
         << "------------------" << endl
         << " ?: Print help" << endl 
         << " f: Toggle full screen mode" << endl
         << " +/-: Increase/Decrease screen point size" << endl 
         << " <drag>+<left button>: rotate model" << endl 
         << " <drag>+<right button>: move model" << endl
         << " <drag>+<middle button>: zoom" << endl
         << " q, <esc>: Quit" << endl << endl; 
}

void usage () {
    printUsage ();
    exit (EXIT_FAILURE);
}



// ------------------------------------

void initLight () {
    GLfloat light_position1[4] = {22.0f, 16.0f, 50.0f, 0.0f};
    GLfloat direction1[3] = {-52.0f,-16.0f,-50.0f};
    GLfloat color1[4] = {0.5f, 1.0f, 0.5f, 1.0f};
    GLfloat ambient[4] = {0.3f, 0.3f, 0.3f, 0.5f};
  
    glLightfv (GL_LIGHT1, GL_POSITION, light_position1);
    glLightfv (GL_LIGHT1, GL_SPOT_DIRECTION, direction1);
    glLightfv (GL_LIGHT1, GL_DIFFUSE, color1);
    glLightfv (GL_LIGHT1, GL_SPECULAR, color1);
    glLightModelfv (GL_LIGHT_MODEL_AMBIENT, ambient);
    glEnable (GL_LIGHT1);
    glEnable (GL_LIGHTING);
}

void init (const char * modelFilename) {
    camera.resize (SCREENWIDTH, SCREENHEIGHT);
    ps.loadPN (modelFilename);
    initLight ();
    glCullFace (GL_BACK);
    glEnable (GL_CULL_FACE);
    glDepthFunc (GL_LESS);
    glEnable (GL_DEPTH_TEST);
    glEnable(GL_POINT_SMOOTH);
    glClearColor (0.0f, 0.0f, 0.0f, 1.0f);
}


// ------------------------------------
// Replace the code of this 
// functions for cleaning memory, 
// closing sockets, etc.
// ------------------------------------

void clear () {
  
}

// ------------------------------------
// Replace the code of this 
// functions for alternative rendering.
// ------------------------------------

void display () {
    glLoadIdentity ();
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    camera.apply ();
    glPointSize (pointSize);
    glCallList (ps.glID);
    glFlush ();
    glutSwapBuffers ();
}

void idle () {
    static float lastTime = glutGet ((GLenum)GLUT_ELAPSED_TIME);
    static unsigned int counter = 0;
    counter++;
    float currentTime = glutGet ((GLenum)GLUT_ELAPSED_TIME);
    if (currentTime - lastTime >= 1000.0f) {
        FPS = counter;
        counter = 0;
        static char winTitle [64];
        unsigned int numOfPoints = ps.V.size ();
        sprintf (winTitle, "gmini - Num. Of Points: %d - FPS: %d", numOfPoints, FPS);
        glutSetWindowTitle (winTitle);
        lastTime = currentTime;
    }
    glutPostRedisplay ();
}

void key (unsigned char keyPressed, int x, int y) {
    Vec3Df l = Vec3Df (50.0, 0.0, 0.0);
    switch (keyPressed) {
    case 'f':
        if (fullScreen == true) {
            glutReshapeWindow (SCREENWIDTH, SCREENHEIGHT);
            fullScreen = false;
        } else {
            glutFullScreen ();
            fullScreen = true;
        }      
        break;
    case '+':
        pointSize += 1.0f;
        break;
    case '-':
        if (pointSize > 1.5f)
            pointSize -= 1.0f;
        break;
    case 'q':
    case 27:
        clear ();
        exit (0);
        break;
    default:
        printUsage ();
        break;
    }
    idle ();
}

void mouse (int button, int state, int x, int y) {
    if (state == GLUT_UP) {
        mouseMovePressed = false;
        mouseRotatePressed = false;
        mouseZoomPressed = false;
    } else {
        if (button == GLUT_LEFT_BUTTON) {
            camera.beginRotate (x, y);
            mouseMovePressed = false;
            mouseRotatePressed = true;
            mouseZoomPressed = false;
        } else if (button == GLUT_RIGHT_BUTTON) {
            lastX = x;
            lastY = y;
            mouseMovePressed = true;
            mouseRotatePressed = false;
            mouseZoomPressed = false;
        } else if (button == GLUT_MIDDLE_BUTTON) {
            if (mouseZoomPressed == false) {
                lastZoom = y;
                mouseMovePressed = false;
                mouseRotatePressed = false;
                mouseZoomPressed = true;
            }
        }
    }
    idle ();
}

void motion (int x, int y) {
    if (mouseRotatePressed == true) {
        camera.rotate (x, y);
    }
    else if (mouseMovePressed == true) {
        camera.move ((x-lastX)/static_cast<float>(SCREENWIDTH), (lastY-y)/static_cast<float>(SCREENHEIGHT), 0.0);
        lastX = x;
        lastY = y;
    }
    else if (mouseZoomPressed == true) {
        camera.zoom (float (y-lastZoom)/SCREENHEIGHT);
        lastZoom = y;
    }
}


void reshape(int w, int h) {
    camera.resize (w, h);
}


int main (int argc, char ** argv) {
    if (argc > 2) {
        printUsage ();
        exit (EXIT_FAILURE);
    }
    glutInit (&argc, argv);
    glutInitDisplayMode (GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
    glutInitWindowSize (SCREENWIDTH, SCREENHEIGHT);
    window = glutCreateWindow ("pMini");
  
    init (argc == 2 ? argv[1] : "african_statue.pn");
    glutIdleFunc (idle);
    glutDisplayFunc (display);
    glutKeyboardFunc (key);
    glutReshapeFunc (reshape);
    glutMotionFunc (motion);
    glutMouseFunc (mouse);
    key ('?', 0, 0);   
    glutMainLoop ();
    return EXIT_SUCCESS;
}

